##########################
##########################
###                    ###
###  MUTE              ###
###  William Woodley   ###
###  24 December 2021  ###
###                    ###
##########################
##########################

# Import packages

import os

import numpy as np
import scipy.integrate as scii

import mute.constants as constants
import mute.surface as surface
import mute.propagation as propagation

# Calculate underground fluxes


def calc_u_fluxes(
    angles=None,
    location="USStd",
    month=None,
    interaction_model="SIBYLL-2.3c",
    primary_model="GSF",
    atmosphere="CORSIKA",
    output=None,
    force=False,
):

    """
    Calculate underground fluxes using surface flux and survival probability matrices stored in /data based on the input parameters and the set global parameters.

    The function searches for the proper surface flux and survival probability files based on the set parameters using the mts.load_s_fluxes_from_file() and mtp.load_survival_probability_tensor_from_file() functions. If a file does not exist, the user is prompted for it to be calculated.

    Parameters
    ----------
    angles : array-like, optional (default: taken from consants.angles)
        An array of zenith angles in degrees to calculate the underground fluxes for.

    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    interaction_model: str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF = GlobalSplineFitBeta
        HG  = HillasGaisser2012 (H3a)
        GH  = GaisserHonda
        ZS  = Zatsepin-Sokolskaya (Default)
        ZSP = Zatsepin-Sokolskaya (PAMELA)
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen")

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    force : bool
        If True, force the calculation of a new surface flux matrix and survival probability tensor if required.

    Returns
    -------
    u_fluxes : tuple of NumPy ndarray
        A two-element tuple. The first element is a two-dimensional array containing the underground fluxes. The second element is a two-dimensional array containing the underground fluxes calculated with an angle of zero degrees for use in calculation of the true vertical intensities.
    """

    # Import packages

    import scipy.interpolate as scii

    # Check values

    if output is None:
        output = constants.get_output()

    # Get the surface flux and survival prbability matrices
    # Getter functions do not take output in as an argument
    # If the calculation functions have to be run through the getter functions, the global output setting is used

    if constants.get_verbose() > 1:
        print("Calculating underground fluxes.")

    s_fluxes = surface.load_s_fluxes_from_file(
        location, month, interaction_model, primary_model, atmosphere, force=force
    )
    survival = propagation.load_survival_probability_tensor_from_file(force=force)

    if (s_fluxes is None or survival is None) and constants.get_verbose() > 1:

        print("Underground fluxes not calculated.")

        return

    # Interpolate the matrices

    interp_at_angles = np.linspace(
        np.min(constants.angles), np.max(constants.angles), 300
    )
    interp_s_fluxes = scii.interp1d(
        constants.ANGLES_FOR_S_FLUXES, s_fluxes, axis=1, kind="cubic"
    )(interp_at_angles)
    interp_s_fluxes = scii.RectBivariateSpline(
        constants.ENERGIES, interp_at_angles, interp_s_fluxes
    )(constants.ENERGIES, constants.angles)

    interp_survival = scii.interp1d(
        constants.SLANT_DEPTHS, survival, axis=1, kind="cubic"
    )(constants.slant_depths)

    # Reshape the matrices

    interp_s_fluxes = np.nan_to_num(
        np.reshape(interp_s_fluxes, (len(constants.ENERGIES), len(constants.angles)))
    )
    interp_survival = np.nan_to_num(
        np.reshape(
            interp_survival,
            (
                len(constants.ENERGIES),
                len(constants.slant_depths),
                len(constants.ENERGIES),
            ),
        )
    )

    # Calculate the underground fluxes for a flat overburden

    if constants.get_overburden() == "flat":

        # Initialise the underground flux matrices
        # First index  = Surface energy
        # Second index = Zenith angle
        # Third index  = Underground energy

        u_fluxes = np.zeros(
            (len(constants.ENERGIES), len(constants.angles), len(constants.ENERGIES))
        )
        u_fluxes_tr = np.zeros(
            (len(constants.ENERGIES), len(constants.angles), len(constants.ENERGIES))
        )

        # Calculate the underground fluxes

        for i in range(len(constants.ENERGIES)):

            for j in range(len(constants.angles)):

                for u in range(len(constants.ENERGIES)):

                    u_fluxes[i, j, u] = (
                        interp_survival[i, j, u]
                        * interp_s_fluxes[i, j]
                        * (constants.E_WIDTHS[i] / constants.E_WIDTHS[u])
                    )
                    u_fluxes_tr[i, j, u] = (
                        interp_survival[i, j, u]
                        * interp_s_fluxes[i, 0]
                        * (constants.E_WIDTHS[i] / constants.E_WIDTHS[u])
                    )

        # Sum over the surface energy grid axis
        # This reduces the number of axes down to two
        # First index = Zenith angle
        # Second index  = Underground energy

        u_fluxes = np.sum(u_fluxes, axis=0)
        u_fluxes_tr = np.sum(u_fluxes_tr, axis=0)

        # Set the angles to default values
        # If user has input singular angle, turn into an array

        if angles is None:
            angles = constants.angles

        angles = np.atleast_1d(angles)

        # If the user has not specified angles, return the whole matrix with no interpolation
        # Check to make sure both arrays have the same number of values, and that those values are equal at each index

        if len(angles) == len(constants.ANGLES) and np.allclose(
            angles, constants.ANGLES
        ):

            if constants.get_verbose() > 1:
                print("Finished calculating underground fluxes.")

            # Write the results to the file

            if output:

                constants.check_directory(
                    os.path.join(constants.get_directory(), "underground"), force=force
                )

                file_name = os.path.join(
                    constants.get_directory(),
                    "underground",
                    "{0}_Underground_Fluxes.txt".format(constants.get_lab()),
                )
                file_out = open(file_name, "w")

                for j in range(len(constants.ANGLES)):

                    for u in range(len(constants.ENERGIES)):

                        file_out.write(
                            "{0:1.5f} {1:1.14f} {2:1.14e}\n".format(
                                angles[j], constants.ENERGIES[u], u_fluxes[j, u]
                            )
                        )

                file_out.close()

                if constants.get_verbose() > 1:
                    print("Underground fluxes written to " + file_name + ".")

            return u_fluxes, u_fluxes_tr

        else:

            # Interpolate at the angles the user has requested

            interp_at_angles = np.linspace(
                np.min(constants.angles), np.max(constants.angles), 300
            )
            interp_u_fluxes = scii.interp1d(
                constants.angles, u_fluxes, axis=0, kind="cubic"
            )(interp_at_angles)
            interp_u_fluxes = scii.RectBivariateSpline(
                interp_at_angles, constants.ENERGIES, interp_u_fluxes
            )(angles, constants.ENERGIES)
            interp_u_fluxes_tr = scii.interp1d(
                constants.angles, u_fluxes_tr, axis=0, kind="cubic"
            )(interp_at_angles)
            interp_u_fluxes_tr = scii.RectBivariateSpline(
                interp_at_angles, constants.ENERGIES, interp_u_fluxes_tr
            )(angles, constants.ENERGIES)

            # Reshape into a matrix of zeroth dimension len(angles)

            interp_u_fluxes = np.nan_to_num(
                np.reshape(interp_u_fluxes, (len(angles), len(constants.ENERGIES)))
            )
            interp_u_fluxes_tr = np.nan_to_num(
                np.reshape(interp_u_fluxes_tr, (len(angles), len(constants.ENERGIES)))
            )

            # Set the global variables

            u_fluxes = interp_u_fluxes
            u_fluxes_tr = interp_u_fluxes_tr

            if constants.get_verbose() > 1:
                print("Finished calculating underground fluxes.")

            # Write the results to the file

            if output:

                constants.check_directory(
                    os.path.join(constants.get_directory(), "underground"), force=force
                )

                file_name = os.path.join(
                    constants.get_directory(),
                    "underground",
                    "{0}_Underground_Fluxes.txt".format(constants.get_lab()),
                )

                file_out = open(file_name, "w")

                for j in range(len(angles)):

                    for u in range(len(constants.ENERGIES)):

                        file_out.write(
                            "{0:1.5f} {1:1.14f} {2:1.14e}\n".format(
                                angles[j], constants.ENERGIES[u], u_fluxes[j, u]
                            )
                        )

                file_out.close()

                if constants.get_verbose() > 1:
                    print("Underground fluxes written to " + file_name + ".")

            return u_fluxes, u_fluxes_tr

    # Calculate the underground fluxes for a non-flat overburden

    else:

        raise NotImplementedError("Non-flat overburdens are not yet implemented.")


# Read the angles and energies from an underground fluxes file


def print_u_fluxes_grids(file_name):

    """Return the zenith angles and underground energies in an underground fluxes file."""

    file_contents = np.loadtxt(constants.get_directory() + "/underground/" + file_name)

    file_angles = np.unique(file_contents[:, 0])
    file_u_energies = np.unique(file_contents[:, 1])

    print("This file has " + str(len(file_angles)) + " zenith angles:")
    print(file_angles)
    print("This file has " + str(len(file_u_energies)) + " underground energies:")
    print(file_u_energies)

    return


# Calculate bins for use in Geant4


def print_geant4_bins(histtype):

    """Print underground energy or zenith angle histogram bins for use in a Geant4 macro with gps."""

    u_fluxes_sum = np.sum(u_fluxes)

    if histtype == "theta":

        u_fluxes_theta_sum = np.sum(u_fluxes, axis=1) / u_fluxes_sum

        print("/gps/hist/type theta")
        print("/gps/hist/point {0:1.5e} {1:1.5e}".format(constants.ANGLES[0], 0))

        for j in range(len(constants.ANGLES)):

            print(
                "/gps/hist/point {0:1.5e} {1:1.5e}".format(
                    constants.ANGLES[j], u_fluxes_theta_sum[j]
                )
            )

    elif histtype == "energy":

        u_fluxes_energy_sum = np.sum(u_fluxes, axis=0) / u_fluxes_sum

        print("/gps/hist/type energy")
        print("/gps/hist/point {0:1.5e} {1:1.5e}".format(constants.E_BINS[0], 0))

        for u in range(len(constants.ENERGIES)):

            print(
                "/gps/hist/point {0:1.5e} {1:1.5e}".format(
                    constants.E_BINS[u + 1], u_fluxes_energy_sum[u]
                )
            )

    else:

        raise ValueError('histtype must be "energy" or "theta".')


# Calculate the underground intensities


def calc_u_intensities(
    angles=None,
    location="USStd",
    month=None,
    interaction_model="SIBYLL-2.3c",
    primary_model="GSF",
    atmosphere="CORSIKA",
    output=None,
    force=False,
):

    """
    Calculate underground intensities.

    Parameters
    ----------
    angles : array-like, optional (default: taken from consants.angles)
        An array of zenith angles in degrees to calculate the underground intensities for.

    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    interaction_model: str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF = GlobalSplineFitBeta
        HG  = HillasGaisser2012 (H3a)
        GH  = GaisserHonda
        ZS  = Zatsepin-Sokolskaya (Default)
        ZSP = Zatsepin-Sokolskaya (PAMELA)
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen")

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    force : bool
        If True, force the calculation of new matrices if required.

    Returns
    -------
    u_intensities : NumPy array
        An array containing the underground intensities.
    """

    # Check values

    if output is None:
        output = constants.get_output()

    # Calculate the underground intensities for a flat overburden

    if constants.get_overburden() == "flat":

        # Set the angles to default values
        # If the user has input a singular angle, turn it into an array

        if angles is None:
            angles = constants.angles

        angles = np.atleast_1d(angles)

        # Calculate the underground fluxes

        u_fluxes = calc_u_fluxes(
            angles,
            location,
            month,
            interaction_model,
            primary_model,
            atmosphere,
            output,
            force=force,
        )

        if u_fluxes is None and constants.get_verbose() > 1:

            print("Underground intensities not calculated.")

            return

        u_fluxes = u_fluxes[0]

        # Initialise the underground intensities array

        u_intensities = np.zeros(len(angles))

        # Calculate the underground intensities

        if constants.get_verbose() > 1:
            print("Calculating underground intensities.")

        for j in range(len(angles)):

            u_intensities[j] = scii.simpson(u_fluxes[j, :], constants.ENERGIES)

        if constants.get_verbose() > 1:
            print("Finished calculating underground intensities.")

        # Write the results to a file

        if output:

            constants.check_directory(
                os.path.join(constants.get_directory(), "underground"), force=force
            )

            file_name = os.path.join(
                constants.get_directory(),
                "underground",
                "{0}_Underground_Intensities.txt".format(constants.get_lab()),
            )

            file_out = open(file_name, "w")

            for j in range(len(angles)):

                file_out.write(
                    "{0:1.5f} {1:1.14e}\n".format(angles[j], u_intensities[j])
                )

            if constants.get_verbose() > 1:
                print("Underground intensities written to " + file_name + ".")

        return u_intensities

    # Calculate the underground intensities for a non-flat overburden

    else:

        raise NotImplementedError("Non-flat overburdens are not yet implemented.")


# Read the angles from an underground intensities file


def print_u_intensities_grid(file_name):

    """Return the zenith angles in an underground intensities file."""

    file_contents = np.loadtxt(constants.get_directory() + "/underground/" + file_name)
    file_angles = np.unique(file_contents[:, 0])

    print("This file has " + str(len(file_angles)) + " zenith angles:")
    print(file_angles)

    return


# Calculate true vertical underground intensities


def calc_u_intensities_tr(
    depths=None,
    location="USStd",
    month=None,
    interaction_model="SIBYLL-2.3c",
    primary_model="GSF",
    atmosphere="CORSIKA",
    output=None,
    force=False,
):

    """
    Calculate true vertical underground intensities.

    Parameters
    ----------
    depths : array-like, optional (default: taken from consants.slant_depths)
        An array of slant depths in [km.w.e.] to calculate the true vertical underground intensities for.

    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    interaction_model: str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF = GlobalSplineFitBeta
        HG  = HillasGaisser2012 (H3a)
        GH  = GaisserHonda
        ZS  = Zatsepin-Sokolskaya (Default)
        ZSP = Zatsepin-Sokolskaya (PAMELA)
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen")

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    force : bool
        If True, force the calculation of new matrices if required.

    Returns
    -------
    u_intensities_tr : NumPy array
        An array containing the true vertical underground intensities.
    """

    # Check values

    if output is None:
        output = constants.get_output()
    if depths is None:
        depths = constants.slant_depths

    depths = np.atleast_1d(depths)

    assert all(
        depths >= np.min(constants.slant_depths)
    ), "Cannot calculate for depths lower than the set vertical depth."
    assert all(depths <= 12), "Cannot calculate for depths greater than 12 km.w.e."

    # Calculate the true vertical underground intensities for a flat overburden

    if constants.get_overburden() == "flat":

        # Calculate the underground fluxes

        u_fluxes = calc_u_fluxes(
            np.degrees(np.arccos(constants.get_vertical_depth() / depths)),
            location,
            month,
            interaction_model,
            primary_model,
            atmosphere,
            output,
            force=force,
        )

        if u_fluxes is None and constants.get_verbose() > 1:

            print("True vertical underground intensities not calculated.")

            return

        u_fluxes_tr = u_fluxes[1]

        # Initialise the underground intensity arrays

        u_intensities_tr = np.zeros(len(depths))

        # Calculate the true vertical underground intensities

        if constants.get_verbose() > 1:
            print("Calculating true vertical underground intensities.")

        for x in range(len(depths)):

            u_intensities_tr[x] = scii.simpson(u_fluxes_tr[x, :], constants.ENERGIES)

        if constants.get_verbose() > 1:
            print("Finished calculating true vertical underground intensities.")

        # Write the results to a file

        if output:

            constants.check_directory(
                os.path.join(constants.get_directory(), "underground"), force=force
            )

            file_name = os.path.join(
                constants.get_directory(),
                "underground",
                "{0}_Underground_Intensities_TR.txt".format(constants.get_lab()),
            )

            file_out = open(file_name, "w")

            for x in range(len(depths)):

                file_out.write(
                    "{0:1.5f} {1:1.14e}\n".format(depths[x], u_intensities_tr[x])
                )

            if constants.get_verbose() > 1:

                print(
                    "True vertical underground intensities written to "
                    + file_name
                    + "."
                )

        return u_intensities_tr

    # Calculate the true vertical underground intensities for a non-flat overburden

    else:

        raise NotImplementedError("Non-flat overburdens are not yet implemented.")


# Read the depths from a true vertical underground intensities file


def print_u_intensities_tr_grid(file_name):

    """Return the slant depths in a true vertical underground intensities file."""

    file_contents = np.loadtxt(constants.get_directory() + "/underground/" + file_name)
    file_depths = np.unique(file_contents[:, 0])

    print("This file has " + str(len(file_depths)) + " slant depths:")
    print(file_depths)

    return


# Calculate vertical-equivalent underground intensities


def calc_u_intensities_eq(
    angles=None,
    location="USStd",
    month=None,
    interaction_model="SIBYLL-2.3c",
    primary_model="GSF",
    atmosphere="CORSIKA",
    output=None,
    force=False,
):

    """
    Calculate vertical-equivalent underground intensities.

    Parameters
    ----------
    angles : array-like, optional (default: taken from consants.angles)
        An array of zenith angles in degrees to calculate the vertical-equivalent underground intensities for.

    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    interaction_model: str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF = GlobalSplineFitBeta
        HG  = HillasGaisser2012 (H3a)
        GH  = GaisserHonda
        ZS  = Zatsepin-Sokolskaya (Default)
        ZSP = Zatsepin-Sokolskaya (PAMELA)
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen")

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    force : bool
        If True, force the calculation of new matrices if required.

    Returns
    -------
    u_intensities_eq : NumPy array
        An array containing the vertical-equivalent underground intensities.
    """

    # Check values

    if output is None:
        output = constants.get_output()

    # Calculate the vertical-equivalent underground intensities for a flat overburden

    if constants.get_overburden() == "flat":

        # Set the angles to the default values
        # If the user has input a singular angle, turn into an array

        if angles is None:
            angles = constants.angles

        angles = np.atleast_1d(angles)

        # Calculate the underground fluxes

        u_fluxes = calc_u_fluxes(
            angles,
            location,
            month,
            interaction_model,
            primary_model,
            atmosphere,
            output,
            force=force,
        )

        if u_fluxes is None and constants.get_verbose() > 1:

            print("Vertical-equivalent underground intensities not calculated.")

            return

        u_fluxes = u_fluxes[0]

        # Initialise the vertical-equivalent underground intensity array

        u_intensities_eq = np.zeros(len(angles))

        # Calculate the vertical-equivalent underground intensities

        if constants.get_verbose() > 1:

            print("Calculating vertical-equivalent underground intensities.")

        for j in range(len(angles)):

            u_intensities_eq[j] = scii.simpson(
                u_fluxes[j, :], constants.ENERGIES
            ) * np.cos(np.radians(angles[j]))

        if constants.get_verbose() > 1:

            print("Finished calculating vertical-equivalent underground intensities.")

        # Write the results to a file

        if output:

            constants.check_directory(
                os.path.join(constants.get_directory(), "underground"), force=force
            )

            file_name = os.path.join(
                constants.get_directory(),
                "underground",
                "{0}_Underground_Intensities_EQ.txt".format(constants.get_lab()),
            )

            file_out = open(file_name, "w")

            for j in range(len(angles)):

                file_out.write(
                    "{0:1.5f} {1:1.14e}\n".format(angles[j], u_intensities_eq[j])
                )

            if constants.get_verbose() > 1:

                print(
                    "Vertical-equivalent underground intensities written to "
                    + file_name
                    + "."
                )

        return u_intensities_eq

    # Calculate the vertical-equivalent underground intensities for a non-flat overburden

    else:

        raise NotImplementedError("Non-flat overburdens are not yet implemented.")


# Read the depths from a vertical-equivalent underground intensities file


def print_u_intensities_eq_grid(file_name):

    """Return the zenith angles in a vertical-equivalent underground intensities file."""

    file_contents = np.loadtxt(constants.get_directory() + "/underground/" + file_name)
    file_angles = np.unique(file_contents[:, 0])

    print("This file has " + str(len(file_angles)) + " zenith angles:")
    print(file_angles)

    return


# Calculate the total fluxes


def calc_u_tot_flux(
    angles=None,
    location="USStd",
    month=None,
    interaction_model="SIBYLL-2.3c",
    primary_model="GSF",
    atmosphere="CORSIKA",
    force=False,
):

    """
    Calculate the total underground flux by calling calc_u_intensities().

    Parameters
    ----------
    angles : array-like, optional (default: taken from consants.angles)
        An array of zenith angles in degrees to calculate the underground intensities for.

    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    interaction_model: str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF = GlobalSplineFitBeta
        HG  = HillasGaisser2012 (H3a)
        GH  = GaisserHonda
        ZS  = Zatsepin-Sokolskaya (Default)
        ZSP = Zatsepin-Sokolskaya (PAMELA)
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen")

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    force : bool
        If True, force the calculation of new matrices if required.

    Returns
    -------
    u_tot_flux : float
        The total underground flux in units of [cm^(-2) s^(-1)].
    """

    # Calculate the total underground flux for a flat overburden

    if constants.get_overburden() == "flat":

        # Set the angles to default values

        if angles is None:

            angles = constants.angles

        # Calculate the underground intensities

        u_intensities = calc_u_intensities(
            angles,
            location,
            month,
            interaction_model,
            primary_model,
            atmosphere,
            force=force,
        )

        if u_intensities is None and constants.get_verbose() > 1:

            print("Total underground fluxes not calculated.")

            return

        # Calculate the total underground flux
        # Because angles goes (0..89), cos(angles) goes (1..0)
        # cos(angles) is decreasing, but scii.simpson() wants an increasing array
        # Therefore, integrate backwards, using [::-1] on the integrand and steps
        # Otherwise, the answer will be negative

        u_tot_flux = (
            2
            * np.pi
            * scii.simpson(u_intensities[::-1], np.cos(np.radians(angles[::-1])))
        )

    # Calculate the total underground flux for a non-flat overburden

    else:

        raise NotImplementedError("Non-flat overburdens are not yet implemented.")

    return u_tot_flux



##########################
##########################
###                    ###
###  MUTE              ###
###  William Woodley   ###
###  19 December 2021  ###
###                    ###
##########################
##########################

# Import packages

import os

import numpy as np
from tqdm import tqdm

import mute.constants as constants

try:

    import proposal as pp

except ImportError:

    pass

# Create the propagator


def _create_propagator(force):

    """This function creates the propagator object in PROPOSAL for use in _propagation_loop()"""

    # Check values

    constants.check_constants(force=force)

    # Global variables
    # The propagator is used in every iteration of the doubly-nested propagation loop
    # Make it a global variable so it only has to be created once

    global propagator

    if constants.get_verbose() > 1:
        print("Creating propagator.")

    # Propagator arguments

    mu = pp.particle.MuMinusDef()
    cuts = pp.EnergyCutSettings(500, 0.05, True)

    if constants.get_medium() == "rock":

        medium = pp.medium.StandardRock()

    elif constants.get_medium() == "water":

        medium = pp.medium.Water()

    elif constants.get_medium() == "ice":

        medium = pp.medium.Ice()

    elif constants.get_medium() == "air":

        medium = pp.medium.Air()

    else:

        raise NotImplementedError(
            "Medium type {0} not implemented.".format(constants.get_medium())
        )

    args = {"particle_def": mu, "target": medium, "interpolate": True, "cuts": cuts}

    # Initialise standard cross-sections, then specify and set parametrisation models

    cross_sections = pp.crosssection.make_std_crosssection(**args)

    brems_param = pp.parametrization.bremsstrahlung.KelnerKokoulinPetrukhin(lpm=False)
    epair_param = pp.parametrization.pairproduction.KelnerKokoulinPetrukhin(lpm=False)
    ionis_param = pp.parametrization.ionization.BetheBlochRossi(energy_cuts=cuts)
    shado_param = pp.parametrization.photonuclear.ShadowButkevichMikheyev()
    photo_param = pp.parametrization.photonuclear.AbramowiczLevinLevyMaor97(
        shadow_effect=shado_param
    )

    cross_sections[0] = pp.crosssection.make_crosssection(brems_param, **args)
    cross_sections[1] = pp.crosssection.make_crosssection(epair_param, **args)
    cross_sections[2] = pp.crosssection.make_crosssection(ionis_param, **args)
    cross_sections[3] = pp.crosssection.make_crosssection(photo_param, **args)

    # Propagation utility

    collection = pp.PropagationUtilityCollection()

    collection.interaction = pp.make_interaction(cross_sections, True)
    collection.displacement = pp.make_displacement(cross_sections, True)
    collection.time = pp.make_time(cross_sections, mu, True)
    collection.decay = pp.make_decay(cross_sections, mu, True)

    pp.PropagationUtilityCollection.cont_rand = False

    utility = pp.PropagationUtility(collection=collection)

    # Other settings

    pp.do_exact_time = False

    # Set up geometry

    detector = pp.geometry.Sphere(
        position=pp.Cartesian3D(0, 0, 0), radius=10000000, inner_radius=0
    )
    density_distr = pp.density_distribution.density_homogeneous(
        mass_density=constants.get_density()
    )

    propagator = pp.Propagator(mu, [(detector, utility, density_distr)])

    if constants.get_verbose() > 1:
        print("Finished creating propagator.")

    return propagator


# Propagation function


def _propagation_loop(energy, slant_depth, force):

    # This function propagates n_muon muons, looping over the energies and slant depths, and returns the muons' underground energies

    # Check values

    n_muon = constants.get_n_muon()

    constants.check_constants(force=force)

    # Convert the slant depth from [km.w.e.] to [cm]

    convert_to_cm = 1e5 * 0.997 / constants.get_density()

    # Initialise the list of underground energies

    u_energies_ix = []

    # Define the initial state of the muon

    mu_initial = pp.particle.ParticleState()
    mu_initial.energy = energy + constants.MU_MASS
    mu_initial.position = pp.Cartesian3D(0, 0, 0)
    mu_initial.direction = pp.Cartesian3D(0, 0, -1)

    # Propagate n_muon muons

    for _ in range(n_muon):

        # Propagate the muons

        track = propagator.propagate(mu_initial, slant_depth * convert_to_cm)

        # Test whether or not the muon has energy left (has not lost all of its energy or has not decayed)
        # If it does, record its energy
        # If it does not, ignore this muon and proceed with the next loop iteration

        if (
            track.track_energies()[-1] != constants.MU_MASS
            and track.track_types()[-1] != pp.particle.Interaction_Type.decay
        ):

            # Store the final underground energy of the muon

            u_energies_ix.append(track.track_energies()[-1])

    # Return the underground energies for the muon

    return u_energies_ix


# Propagate the muons and return underground energies


def propagate_muons(seed=0, job_array_number=0, output=None, force=False):

    """
    Propagate muons for the default surface energy grid and slant depths.

    The default surface energy grid is given by constants.ENERGIES, and the default slant depths are given by constants.SLANT_DEPTHS.

    Parameters
    ----------
    seed : str, optional (default: 0)
        The random seed for use in the PROPOSAL propagator.

    job_array_number : int, optional (default: 0)
        The job array number from a high-statistics run on a computer cluster. This is set so the underground energy files from each job in the job array will be named differently.

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    force : bool, optional (default: False)
        If True, this will force the creation of an underground_energies directory if one does not already exist.

    Returns
    -------
    u_energies : NumPy ndarray
        A two-dimensional array containing lists of underground energies for muons that survived the propagation.
    """

    # Check values

    assert type(job_array_number) == int, "job_array_number must be an integer."

    constants.check_constants(force=force)

    if output is None:
        output = constants.get_output()

    # Create the propagator once

    _create_propagator(force=force)

    # Set the random seed

    pp.RandomGenerator.get().set_seed(seed)

    # Initialise the matrix of underground energies

    u_energies = np.zeros(
        (len(constants.ENERGIES), len(constants.SLANT_DEPTHS)), dtype=np.ndarray
    )

    # Run the propagation function and print the underground energies

    if constants.get_verbose() >= 1:
        print(
            "Propagating "
            + str(
                constants.get_n_muon()
                * len(constants.ENERGIES)
                * len(constants.SLANT_DEPTHS)
            )
            + " muons."
        )

    for i in (
        tqdm(range(len(constants.ENERGIES)))
        if constants.get_verbose() >= 1
        else range(len(constants.ENERGIES))
    ):

        for x in range(len(constants.SLANT_DEPTHS)):

            u_energies[i, x] = _propagation_loop(
                constants.ENERGIES[i], constants.SLANT_DEPTHS[x], force=force
            )

    if constants.get_verbose() >= 1:
        print("Finished propagation.")

    if output:

        constants.check_directory(
            os.path.join(constants.get_directory(), "underground_energies"), force=force
        )

        file_name = os.path.join(
            constants.get_directory(),
            "underground_energies",
            "{0}_{1}_{2}_Underground_Energies_{3}.npy".format(
                constants.get_medium(),
                constants.get_density(),
                constants.get_n_muon(),
                job_array_number,
            ),
        )

        np.save(file_name, u_energies)

        if constants.get_verbose() > 1:
            print("Underground energies written to " + file_name + ".")

    return u_energies


# Load underground energies


def _load_u_energies_from_files(file_name, n_job=1, force=False):

    """
    Load the underground energies resulting from the PROPOSAL Monte Carlo from a file or collection of files stored in data/underground_energies.

    Parameters
    ----------
    file_name : str, optional
        The file name pattern that the underground energy data is stored in. This should end in an underscore so the function can append the job array number.

    n_job : int, optional (default: 1)
        The number of jobs that were run on the computer cluster. Set this to the number of files the underground energies are spread across.

    force : bool, optional (default: False)
        If True, this will force the creation of an underground_energies directory if one does not already exist.

    Returns
    -------
    u_energies : NumPy ndarray
        A two-dimensional array containing lists of underground energies for muons that survived the propagation.
    """

    # Check values

    constants.check_constants(force=force)

    # Check that the directory exists

    if not os.path.exists(constants.get_directory() + "/underground_energies"):

        if constants.get_verbose() >= 1:

            print(
                constants.get_directory()
                + "/underground_energies does not exist. Underground energies not loaded."
            )

        return None

    # Test if the file exists

    if not os.path.isfile(file_name + "_0.npy"):

        if constants.get_verbose() >= 1:

            print(file_name + "_0.npy does not exist. Underground energies not loaded.")

        return None

    # Fill a u_energies array with empty lists that will be able to be extended

    u_energies = np.empty(
        (len(constants.ENERGIES), len(constants.SLANT_DEPTHS)), dtype=object
    )

    for i in np.ndindex(u_energies.shape):
        u_energies[i] = []

    # Loop over all output files and add the contents to u_energies

    for a in tqdm(range(n_job)) if constants.get_verbose() >= 1 else range(n_job):

        u_energies += np.load(file_name + "_" + str(a) + ".npy", allow_pickle=True)

    if constants.get_verbose() > 1:
        print("Loaded underground energies.")

    return u_energies


def calc_survival_probability_tensor(
    seed=0, file_name=None, n_job=1, output=None, force=False
):

    """
    Calculate survival probabilities for the default surface energy grid and slant depths.

    The default surface energy grid is given by constants.ENERGIES, and the default slant depths are given by constants.SLANT_DEPTHS. If the propagation of muons has already been done, this will load the underground energies file (it will load it for n_job = 1; to load for more jobs, call load_u_energies_from_files() directly), unless force is set to True.

    Parameters
    ----------
    seed : int, optional (default: 0)
        The random seed for use in the PROPOSAL propagator.

    file_name : str, optional
        The file name pattern that the underground energy data is stored in. This should end in an underscore so the function can append the job array number.

    n_job : int, optional (default: 1)
        The number of jobs that were run on the computer cluster. Set this to the number of files the underground energies are spread across.

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    force : bool, optional (default: False)
        If True, this will force the muons to be propagated whether an underground energies file already exists or not.

    Returns
    -------
    survival : NumPy ndarray
        A three-dimensional array containing the survival probabilities.
    """

    # Check values

    if output is None:
        output = constants.get_output()

    # Construct a file name

    file_name_default = os.path.join(
        constants.get_directory(),
        "underground_energies",
        "{0}_{1}_{2}_Underground_Energies".format(
            constants.get_medium(),
            constants.get_density(),
            int(constants.get_n_muon() / n_job),
        ),
    )

    # Check if propagate_muons() should be forced or not
    # If not, check if underground energy files exist
    # If not, ask if muons should be propagated

    if force:

        u_energies = propagate_muons(seed=seed, output=output, force=force)

    else:

        # Check if the user has specified underground energies to load
        # If not, look for the default file name pattern, and check if it exists
        # If so, load the underground energies
        # If not, ask if muons should be propagated

        if file_name is not None:

            u_energies = _load_u_energies_from_files(
                file_name=os.path.join(
                    constants.get_directory(), "underground_energies", file_name
                ),
                n_job=n_job,
                force=force,
            )

        elif os.path.isfile(file_name_default + "_0.npy"):

            u_energies = _load_u_energies_from_files(
                file_name=file_name_default, n_job=n_job, force=force
            )

        else:

            answer = input(
                "No underground energy file currently exists for the set lab, medium, or number of muons. Would you like to create one (y/n)?: "
            )

            if answer.lower() == "y":

                u_energies = propagate_muons(seed=seed, output=output, force=force)

            else:

                print("Underground energies not calculated.")
                print("Survival probabilities not calculated.")

                return

    # Check that the underground energies were loaded

    if u_energies is None:

        print("Survival probabilities not calculated.")

        return

    # Calculate the survival probabilities
    # First index  = Surface energy
    # Second index = Slant depth
    # Third index  = Underground energy

    survival = np.zeros(
        (len(constants.ENERGIES), len(constants.SLANT_DEPTHS), len(constants.ENERGIES))
    )

    if constants.get_verbose() > 1:
        print("Calculating survival probabilities.")

    for i in range(len(constants.ENERGIES)):

        for x in range(len(constants.SLANT_DEPTHS)):

            survival[i, x, :] = np.histogram(
                np.array(u_energies[i, x]), bins=constants.E_BINS
            )[0] / float(constants.get_n_muon())

    if constants.get_verbose() > 1:
        print("Finished calculating survival probabilities.")

    # Write the results to a file

    if output:

        constants.check_directory(
            os.path.join(constants.get_directory(), "survival_probabilities"),
            force=force,
        )

        file_name = os.path.join(
            constants.get_directory(),
            "survival_probabilities",
            "{0}_{1}_{2}_Survival_Probabilities.txt".format(
                constants.get_medium(), constants.get_density(), constants.get_n_muon()
            ),
        )

        file_out = open(file_name, "w")

        for i in range(len(constants.ENERGIES)):

            for x in range(len(constants.SLANT_DEPTHS)):

                for u in range(len(constants.ENERGIES)):

                    file_out.write(
                        "{0:1.14f} {1:1.5f} {2:1.14f} {3:1.14e}\n".format(
                            constants.ENERGIES[i],
                            constants.SLANT_DEPTHS[x],
                            constants.ENERGIES[u],
                            survival[i, x, u],
                        )
                    )

        file_out.close()

        if constants.get_verbose() > 1:
            print("Survival probabilities written to " + file_name + ".")

    return survival


def load_survival_probability_tensor_from_file(force=False):

    """
    Retrieve a survival probability matrix stored in data/survival_probabilities based on the set global parameters.

    The function searches for a file name that matches the set lab, medium, and number of muons. If the file does not exist, prompt the user to run calc_survival().

    Parameters
    ----------
    force : bool
        If True, force the calculation of a new survival probability tensor if required.

    Returns
    -------
    survival : NumPy ndarray
        A two-dimensional array containing the survival probabilities.
    """

    # Define a function to run if there is no survival probability file

    def no_file(force):

        # If the file does not exist, ask the user if they want to run PROPOSAL to create it

        if not force:

            answer = input(
                "No survival probability matrix currently exists for the set lab, medium, or number of muons. Would you like to create one (y/n)?: "
            )

        if force or answer.lower() == "y":

            survival_full = calc_survival_probability_tensor(force=force)

            return survival_full

        else:

            print("Survival probabilities not calculated.")

            return None

    # Construct a file name based on the set lab, medium, and number of muons

    file_name = os.path.join(
        constants.get_directory(),
        "survival_probabilities",
        "{0}_{1}_{2}_Survival_Probabilities.txt".format(
            constants.get_medium(), constants.get_density(), constants.get_n_muon()
        ),
    )

    # Check if the file exists
    # Also check that the file has the correct number of energies and angles

    if os.path.isfile(file_name):

        if constants.get_verbose() > 1:
            print("Loading survival probabilities from " + file_name + ".")

        # If the file exists, read in the survival probabilities from it

        file = open(file_name, "r")
        n_lines = len(file.read().splitlines())

        file.close()

        if n_lines == constants.len_iju:

            survival = np.reshape(
                np.loadtxt(file_name)[:, 3],
                (
                    len(constants.ENERGIES),
                    len(constants.SLANT_DEPTHS),
                    len(constants.ENERGIES),
                ),
            )

            if constants.get_verbose() > 1:
                print("Loaded survival probabilities.")

            return survival

        else:

            return no_file(force=force)

    else:

        return no_file(force=force)


# Read the energies and slant depths from a survival probabilities file


def print_survival_probability_tensor_grids(file_name):

    """Return the slant depths and underground energies in a survival probability file."""

    file_contents = np.loadtxt(
        os.path.join(constants.get_directory(), "survival_probabilities", file_name)
    )

    file_s_energies = np.unique(file_contents[:, 0])
    file_slant_depths = np.unique(file_contents[:, 1])
    file_u_energies = np.unique(file_contents[:, 2])

    print("This file has " + str(len(file_s_energies)) + " surface energies:")
    print(file_s_energies)
    print("This file has " + str(len(file_slant_depths)) + " slant depths:")
    print(file_slant_depths)
    print("This file has " + str(len(file_u_energies)) + " underground energies:")
    print(file_u_energies)

    return



import numpy as np

import mute.constants as mtc
import mute.surface as mts


def test_s_fluxes():

    mtc.clear()
    mtc.set_directory(".")

    s_fluxes_calc = mts.calc_s_fluxes(output=False, test=True)
    s_fluxes_read = mts.load_s_fluxes_from_file(test=True)

    assert np.allclose(s_fluxes_calc, s_fluxes_read)



import os

import numpy as np

import mute.constants as mtc
import mute.underground as mtu


def test_u_intensities():

    mtc.clear()

    mtc.set_output(False)
    mtc.set_vertical_depth(3.5)

    u_tot_flux_calc = mtu.calc_u_tot_flux(force=True)
    u_tot_flux_read = 1.1227419181425535e-08

    assert np.allclose(u_tot_flux_calc, u_tot_flux_read)



import os


def download_file(url, dir_out):

    """Download the data files from GitHub."""

    import requests
    from zipfile import ZipFile

    # Download the zip file

    response = requests.get(url, stream=True)
    file_out = "mute_data_files.zip"

    with open(file_out, "wb") as f:

        for line in response.iter_content(chunk_size=1024 * 1024):

            f.write(line)

    # Unzip the file

    with ZipFile(file_out, "r") as f:

        f.extractall(dir_out)

    os.chdir(dir_out)
    os.rename("data_20211219", "data")
    os.chdir("..")

    # Delete the zip file

    os.remove(file_out)


# Download the files from GitHub

if not os.path.isfile(
    os.path.join(os.path.dirname(__file__), "data", "data_20211219.txt")
):

    download_file(
        "https://github.com/wjwoodley/mute/releases/download/0.1.0/data_20211219.zip",
        os.path.dirname(__file__),
    )



##########################
##########################
###                    ###
###  MUTE              ###
###  William Woodley   ###
###  24 December 2021  ###
###                    ###
##########################
##########################

# Import packages

import os

import numpy as np

# Energies in [MeV]

E_BINS = np.logspace(1.9, 14, 122)[:-30]  # Bin edges
E_WIDTHS = np.diff(E_BINS)  # Bin widths
ENERGIES = np.sqrt(E_BINS[1:] * E_BINS[:-1])  # Bin centers

e_bins = E_BINS
e_widths = E_WIDTHS
energies = ENERGIES

# Slant depths in [km.w.e.] and angles in [degrees]
# These are the defaults used to construct the matrices
# The user can enter their own angles in the function calls, which will interpolate over the grids created by these angles

SLANT_DEPTHS = np.linspace(1, 12, 23)
ANGLES = np.degrees(np.arccos(1 / SLANT_DEPTHS))

slant_depths = SLANT_DEPTHS
angles = ANGLES

# Angles in [degrees] specifically for the calculation of surface flux matrices

ANGLES_FOR_S_FLUXES = np.linspace(0, 90, 20)

# Length values for file comparisons

len_ij = len(ENERGIES) * len(ANGLES_FOR_S_FLUXES)
len_iju = len(ENERGIES) * len(SLANT_DEPTHS) * len(ENERGIES)

# The rest mass of a muon in [MeV]

MU_MASS = 105.6583745

# Define user-set constants

_verbose = 2
_output = True
_directory = os.path.join(os.path.dirname(__file__), "data")
_lab = "Default"
_overburden = "flat"
_vertical_depth = 1
_medium = "rock"
_density = 2.65
_n_muon = 100000

# Setters and getters for global constants

# Verbose


def set_verbose(verbose):

    """
    Set the verbosity level.

    Parameters
    ----------
    verbose : int (default: 2)
        The verbosity level. Options:
        0 = Print no information.
        1 = Print the progress through the propagation loop if propagation muons or the loading of the Monte Carlo underground energies from files.
        2 = Print when calculations start and finish, and where data is read from and written to.
    """

    assert type(verbose) is int

    global _verbose

    _verbose = verbose


def get_verbose():

    """Return the set verbosity level."""

    return _verbose


# Whether or not the results are written to output files in directory


def set_output(output):

    """
    Set whether results are output to files or not. The default is True.

    Most functions also have their own optional output keyword argument to give more control over which output files are created.
    """

    assert type(output) is bool

    global _output

    _output = output


def get_output():

    """Return the set output setting."""

    return _output


# Directory for input and output


def set_directory(directory):

    """Set the working directory for files to be written to and read from. The default is \"data\"."""

    from .__init__ import download_file

    global _directory

    _directory = directory

    if not os.path.isfile(os.path.join(_directory, "data_20211219.txt")):

        download_file(
            "https://github.com/wjwoodley/mute/releases/download/0.1.0/data_20211219.zip",
            _directory,
        )


def get_directory():

    """Return the set working directory for files to be written to and read from."""

    return _directory


# Lab name used in output files


def set_lab(lab):

    """Set the name of the lab. This is used in the output file names as a way to differentiate different data files. The default is \"Default\"."""

    global _lab

    _lab = lab


def get_lab():

    """Return the set name of the lab."""

    return _lab


# Overburden


def set_overburden(overburden):

    """Set the overburden type. The default is  \"flat\" overburden."""

    assert overburden in [
        "flat",
        "mountain",
    ], 'overburden must be set to either "flat" or "mountain".'

    global _overburden

    _overburden = overburden


def get_overburden():

    """Return the set overburden type."""

    return _overburden


# Vertical depth in [km.w.e.]


def set_vertical_depth(vertical_depth):

    """Set the vertical depth, h. The default is 1 km.w.e."""

    global _vertical_depth
    global slant_depths
    global angles

    _vertical_depth = vertical_depth

    # Use the set vertical depth to calculate new slant depths and zenith angles

    slant_depths = np.sort(
        np.concatenate(
            ([_vertical_depth], SLANT_DEPTHS[SLANT_DEPTHS > _vertical_depth])
        )
    )
    angles = np.degrees(np.arccos(_vertical_depth / slant_depths))


def get_vertical_depth():

    """Return the set vertical depth."""

    return _vertical_depth


# Medium


def set_medium(medium):

    """Set the medium for the muons to be propagated through. The default is standard rock (\"rock\")."""

    assert medium in [
        "rock",
        "water",
        "ice",
        "air",
    ], 'medium must be set to "rock", "water", "ice", or "air".'

    global _medium

    _medium = medium


def get_medium():

    """Return the set propagation medium."""

    return _medium


# Density in [gcm^-3]


def set_density(density):

    """Set the density of the propagation medium. The default is 2.65 g/cm^3 (the density of standard rock)."""

    assert density > 0, "Media density must be positive."

    global _density

    _density = density


def get_density():

    """Return the set density of the propagation medium."""

    return _density


# Number of muon


def set_n_muon(n_muon):

    """Set the number of muons to be propagated per surface energy-slant depth bin. The default is 10."""

    assert type(n_muon) is int, "Number of muons must be an integer."
    assert n_muon > 0, "Number of muons must be positive."

    global _n_muon

    _n_muon = n_muon


def get_n_muon():

    """Return the set number of muons to be propagated per surface energy-slant depth bin."""

    return _n_muon


# Check if required directories exist


def check_directory(directory, force=False):

    """Check whether a directory required to store output files exists or not. If it does not, ask the user if it should be created."""

    if not os.path.exists(directory):

        # If the directory does not exist, ask the user if they want to create it

        if not force:

            answer = input(
                directory
                + " does not currently exist. Would you like to create it (y/n)?: "
            )

        if force or answer.lower() == "y":

            os.makedirs(directory)

        else:

            print("Directory not created.")

            return

    else:

        return


# Check that the constants are set to correctly before running any calculation functions


def check_constants(force=False):

    """Check that the constants are set correctly before running any calculation functions."""

    # Check that the working directory the user has set exists

    check_directory(get_directory(), force=force)

    # Check that the overburden has been set to one of the available options

    if _overburden is None:

        raise ValueError('Overburden must be set to either "flat" or "mountain".')

    # Check that the vertical depth is set correctly, if needed

    elif _overburden == "flat":

        assert _vertical_depth is not None, "Initial vertical depth must be set."
        assert _vertical_depth > 0, "Initial vertical depth must be positive."

    # Check that the number of muons is set correctly

    if _n_muon <= 0:

        raise ValueError("Number of muons must be a positive integer.")


# Calculate weights


def calc_weights(
    file, units="kmwe", n_bins=[10, 10], max_slant_depth=None, force=False
):

    """
    Calculate the normalised weight matrix used for non-flat overburdens.

    Parameters
    ----------
    file : str
        The full path and name of the .txt file containing the geometry of the mountain. The first column should be the zenith angle in degrees; the second column should be the azimuthal angle in degrees; the third column should be the slant depth in units of units (see below).

    units : str, optional (default: "kmwe")
        The units of the slant depths in file. This must be one of ["m", "km", "mwe", "kmwe"].

    n_bins : array-like, optional (default: [10, 10])
        The number of bins in cos(theta) and slant depth, respectively

    max_slant_depth : float, optional (default: None)
        The maximum slant depth to use when creating the weights. Any data for slant depths above this value will be excluded. If None, all slant depths will be used.

    force : bool
        If True, this will force the creation of a directory to store output if one does not already exist.

    Returns
    -------
    cos_theta : NumPy array
        The centers of the cosine of the zenith angle bins of the weights histogram. The length is equal to n_bins[0].

    slant_depths : NumPy array
        The centers of the slant depth bins of the weights histogram. The length is equal to n_bins[1].

    weights : NumPy ndarray
        A two-dimensional array of shape (len(cos_theta), len(mu_slants)) containing the bin values of the normalised weights histogram.
    """

    # Check values

    assert (
        overburden == "mountain"
    ), "Weights can only be calculated for mountain overburdens."
    assert units_in in ["m", "km", "mwe", "kmwe"], "Units must be m, km, mwe, or kmwe."

    check_constants(force=force)

    # Global variables

    global mountain_zenith
    global mountain_azimuthal
    global mountain_slants

    global weights
    global cos_bin_width
    global slant_bin_width

    global slant_depths
    global angles

    # If the data is in [m] or [km], convert to [m.w.e.] or [km.w.e.]

    density_mult = 1

    if units == "m" or units == "km":
        density_mult = density / 0.997

    # If the data is in [m] or [m.w.e.], convert to [km.w.e.]

    scale = 1

    if units == "m" or units == "mwe":
        scale = 1e-3

    # Load the mountain geometry data from the file

    mountain_zenith_all = np.loadtxt(file)[:, 0]
    mountain_azimuthal_all = np.loadtxt(file)[:, 1]
    mountain_slants_all = np.loadtxt(file)[:, 2] * density_mult * scale

    mountain_zenith = np.unique(mountain_zenith_all)
    mountain_azimuthal = np.unique(mountain_azimuthal_all)
    mountain_slants = np.reshape(
        mountain_slants_all, (len(mountain_zenith), len(mountain_azimuthal))
    )

    # Define bins edges from the data

    if max_slant is None:

        slant_bins = np.sort(np.histogram(mountain_slants_all, bins=n_bins[1])[1])
        cos_bins = np.linspace(
            np.cos(np.radians(max(mountain_zenith_all))), 1, n_bins[0] + 1
        )

    else:

        sel = np.where(np.array(mountain_slants_all) < max_slant_in)

        slant_bins = np.sort(np.histogram(mountain_slants_all[sel], bins=n_bins[1])[1])
        cos_bins = np.linspace(
            np.cos(np.radians(max(mountain_zenith_all[sel]))), 1, n_bins[0] + 1
        )

    # Calculate the weights

    weights = np.histogram2d(
        np.cos(np.radians(mountain_zenith_all)),
        mountain_slants_all,
        bins=[cos_bins, slant_bins],
        density=True,
    )[0]

    # Calculate the bin widths

    slant_bin_width = np.diff(slant_bins)[0]
    cos_bin_width = np.diff(cos_bins)[0]

    # Calculate the bin centers

    slant_depths = (slant_bins[1:] + slant_bins[:-1]) / 2
    angles = np.degrees(np.arccos((cos_bins[1:] + cos_bins[:-1]) / 2))

    return cos_theta, slant_depths, weights


# Clear all variables


def clear():

    """Reset all of the values set or calculated in MUTE to their default values."""

    # Import packages

    import gc

    # Global constants and variables

    global _verbose
    global _output
    global _directory
    global _lab
    global _overburden
    global _vertical_depth
    global _medium
    global _density
    global _n_muon

    _verbose = 2
    _output = True
    _directory = os.path.join(os.path.dirname(__file__), "data")
    _lab = "Default"
    _overburden = "flat"
    _vertical_depth = 1
    _medium = "rock"
    _density = 2.65
    _n_muon = 100000

    # Global weight variables
    # Set the variables to None first, in case they do not already exist

    global mountain_zenith
    global mountain_azimuthal
    global mountain_slants
    global weights
    global cos_bin_width
    global slant_bin_width

    mountain_zenith = None
    mountain_azimuthal = None
    mountain_slants = None
    weights = None
    cos_bin_width = None
    slant_bin_width = None

    del mountain_zenith
    del mountain_azimuthal
    del mountain_slants
    del weights
    del cos_bin_width
    del slant_bin_width

    # Energies, slant depths, zenith angles, and the muon mass

    global E_BINS
    global E_WIDTHS
    global ENERGIES
    global e_bins
    global e_widths
    global energies
    global SLANT_DEPTHS
    global ANGLES
    global slant_depths
    global angles
    global ANGLES_FOR_S_FLUXES
    global len_ij
    global len_iju
    global MU_MASS

    E_BINS = np.logspace(1.9, 14, 122)[:-30]  # [MeV]
    E_WIDTHS = np.diff(E_BINS)  # [MeV]
    ENERGIES = np.sqrt(E_BINS[1:] * E_BINS[:-1])  # [MeV]
    e_bins = E_BINS  # [MeV]
    e_widths = E_WIDTHS  # [MeV]
    energies = ENERGIES  # [MeV]
    SLANT_DEPTHS = np.linspace(1, 12, 23)  # [km.w.e.]
    ANGLES = np.degrees(np.arccos(1 / SLANT_DEPTHS))  # [degrees]
    slant_depths = SLANT_DEPTHS  # [km.w.e.]
    angles = ANGLES  # [degrees]
    ANGLES_FOR_S_FLUXES = np.linspace(0, 90, 20)  # [degrees]
    len_ij = len(ENERGIES) * len(ANGLES_FOR_S_FLUXES)
    len_iju = len(ENERGIES) * len(SLANT_DEPTHS) * len(ENERGIES)
    MU_MASS = 105.6583745  # [MeV]

    gc.collect()



##########################
##########################
###                    ###
###  MUTE              ###
###  William Woodley   ###
###  19 December 2021  ###
###                    ###
##########################
##########################

# Import packages

import os

import numpy as np
from tqdm import tqdm

import mute.constants as constants

# Calculate surface fluxes


def calc_s_fluxes(
    location="USStd",
    month=None,
    interaction_model="SIBYLL-2.3c",
    primary_model="GSF",
    atmosphere="CORSIKA",
    output=None,
    force=False,
    test=False,
):

    """
    Calculate surface fluxes for default surface energy grid and zenith angles.

    The default surface energy grid are given by constants.ENERGIES, and the default zenith angles are given by constants.ANGLES_FOR_S_FLUXES.

    Parameters
    ----------
    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    interaction_model: str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF = GlobalSplineFitBeta
        HG  = HillasGaisser2012 (H3a)
        GH  = GaisserHonda
        ZS  = Zatsepin-Sokolskaya (Default)
        ZSP = Zatsepin-Sokolskaya (PAMELA)
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen")

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    output : bool, optional (default: taken from constants.get_output())
        If True, an output file will be created to store the results.

    force : bool, optional (default: False)
        If True, this will force the creation of a surface_fluxes directory if one does not already exist.

    test: bool, optional (default: False)
        For use in the file test_s_fluxes.py. If True, this will calculate surface fluxes for only three angles.

    Returns
    -------
    s_fluxes : NumPy ndarray
        A two-dimensional array containing the surface fluxes.
    """

    # Import MCEq

    from MCEq.core import MCEqRun
    import mceq_config
    import crflux.models as pm

    # Check values

    constants.check_constants(force=force)

    if output is None:
        output = constants.get_output()

    primary_models = {
        "GSF": (pm.GlobalSplineFitBeta, None),
        "HG": (pm.HillasGaisser2012, "H3a"),
        "GH": (pm.GaisserHonda, None),
        "ZS": (pm.ZatsepinSokolskaya, "default"),
        "ZSP": (pm.ZatsepinSokolskaya, "pamela"),
    }

    if isinstance(primary_model, str):

        assert primary_model in primary_models, "Set primary model not available."

        primary_model_for_MCEq = primary_models[primary_model]
        pm_sname = primary_model

    elif isinstance(primary_model, tuple) and len(primary_model) == 2:

        primary_model_for_MCEq = primary_model
        pm_sname = primary_model[0](primary_model[1]).sname

    else:

        raise TypeError("Primary model not set correctly.")

    assert atmosphere in ["CORSIKA", "MSIS00"]

    angles = constants.ANGLES_FOR_S_FLUXES

    if test:

        angles = [0, 30, 60]

    # Set MCEq up

    if constants.get_verbose() > 1:

        print(
            "Calculating surface fluxes for "
            + location
            + " using "
            + interaction_model
            + " and "
            + pm_sname
            + "."
        )

        mceq_run = MCEqRun(
            interaction_model=interaction_model,
            primary_model=primary_model_for_MCEq,
            theta_deg=0.0,
            **mceq_config.config
        )
        mceq_run.set_density_model((atmosphere, (location, month)))

    else:

        from contextlib import redirect_stdout

        with open(os.devnull, "w") as suppress, redirect_stdout(suppress):

            mceq_run = MCEqRun(
                interaction_model=interaction_model,
                primary_model=primary_model_for_MCEq,
                theta_deg=0.0,
                **mceq_config.config
            )
            mceq_run.set_density_model((atmosphere, (location, month)))

    # Run MCEq
    # Convert the surface fluxes from [GeV] to [MeV] to use in PROPOSAL
    # First index  = Surface energy
    # Second index = Zenith angle

    s_fluxes = np.zeros((len(constants.ENERGIES), len(angles)))

    # Run MCEq

    for j in (
        tqdm(range(len(angles))) if constants.get_verbose() >= 1 else range(len(angles))
    ):

        mceq_run.set_theta_deg(angles[j])
        mceq_run.solve()

        s_fluxes[:, j] = (
            1e-3
            * (
                mceq_run.get_solution("total_mu+", mag=0)
                + mceq_run.get_solution("total_mu-", mag=0)
            )[:-30]
        )

    if constants.get_verbose() > 1:
        print("Finished calculating surface fluxes.")

    # Write the results to a file

    if output:

        constants.check_directory(
            os.path.join(constants.get_directory(), "surface_fluxes"), force=force
        )

        if month is None:

            file_name = os.path.join(
                constants.get_directory(),
                "surface_fluxes",
                "Surface_Fluxes_{0}_{1}_{2}.txt".format(
                    location, interaction_model, pm_sname
                ),
            )

        else:

            file_name = os.path.join(
                constants.get_directory(),
                "surface_fluxes",
                "Surface_Fluxes_{0}_{1}_{2}_{3}.txt".format(
                    location, month, interaction_model, pm_sname
                ),
            )

        file_out = open(file_name, "w")

        for i in range(len(constants.ENERGIES)):

            for j in range(len(angles)):

                file_out.write(
                    "{0:1.14f} {1:1.5f} {2:1.14e}\n".format(
                        constants.ENERGIES[i],
                        angles[j],
                        s_fluxes[i, j],
                    )
                )

        file_out.close()

        if constants.get_verbose() > 1:
            print("Surface fluxes written to " + file_name + ".")

    return s_fluxes


# Get surface fluxes


def load_s_fluxes_from_file(
    location="USStd",
    month=None,
    interaction_model="SIBYLL-2.3c",
    primary_model="GSF",
    atmosphere="CORSIKA",
    force=False,
    test=False,
):

    """
    Retrieve a surface fluxes matrix stored in data/surface_fluxes based on the input parameters.

    The function searches for a file name that matches the set location, month, interaction model, and primary model. If the file does not exist, prompt the user to run calc_s_fluxes().

    Parameters
    ----------
    location : str, optional (default: "USStd")
        The name of the location for which to calculate the surface fluxes. See the Tutorial or MCEq documentation for a list of options.

    month : str, optional (default: None)
        The month for which to calculate the surface fluxes. For US Standard Atmosphere, use None. For seasonal variations, use the month name.

    interaction_model: str, optional (default: "SIBYLL-2.3c")
        The hadronic interaction model to use in MCEq. See the Tutorial or MCEq documentation for a list of options.

    primary_model : str in {"GSF", "HG", "GH", "ZS", "ZSP"} or tuple, optional (default: "GSF")
        The primary flux model to use in MCEq. Options:
        GSF = GlobalSplineFitBeta
        HG  = HillasGaisser2012 (H3a)
        GH  = GaisserHonda
        ZS  = Zatsepin-Sokolskaya (Default)
        ZSP = Zatsepin-Sokolskaya (PAMELA)
        Alternatively, this can be set with a tuple. For example: (pm.GaisserStanevTilav, "3-gen")

    atmosphere : {"CORSIKA", "MSIS00"}, optional (default: "CORSIKA")
        The atmospheric model. For US Standard Atmosphere, use "CORSIKA". For seasonal variations, use "MSIS00".

    force : bool, optional (default: False)
        If True, force the calculation of a new surface fluxes matrix if required.

    test: bool, optional (default: False)
        For use in the file test_s_fluxes.py. If True, this will load surface fluxes for only three angles.

    Returns
    -------
    s_fluxes : NumPy ndarray
        A two-dimensional array containing the surface fluxes.
    """

    # Check values

    constants.check_constants(force=force)

    angles = constants.ANGLES_FOR_S_FLUXES

    if test:

        angles = [0, 30, 60]

    # Define a function to run if there is no surface fluxes file

    def no_file(force):

        # If the file does not exist, ask the user if they want to run MCEq to create it

        if not force:

            answer = input(
                "No surface flux matrix currently exists for these models. Would you like to create one (y/n)?: "
            )

        if force or answer.lower() == "y":

            s_fluxes_full = calc_s_fluxes(
                location,
                month,
                interaction_model,
                primary_model,
                atmosphere,
                force=force,
            )

            return s_fluxes_full

        else:

            print("Surface fluxes not calculated.")

            return None

    # Construct the file name based on the user's inputs

    if month is None:

        file_name = os.path.join(
            constants.get_directory(),
            "surface_fluxes",
            "Surface_Fluxes_{0}_{1}_{2}.txt".format(
                location, interaction_model, primary_model
            ),
        )

    else:

        file_name = os.path.join(
            constants.get_directory(),
            "surface_fluxes",
            "Surface_Fluxes_{0}_{1}_{2}_{3}.txt".format(
                location, month, interaction_model, primary_model
            ),
        )

    # Check if the file exists

    if os.path.exists(file_name):

        if constants.get_verbose() > 1:
            print(
                "Loading surface fluxes for "
                + location
                + " using "
                + interaction_model
                + " and "
                + primary_model
                + "."
            )

        # If the file exists, read the fluxes in from it

        file = open(file_name, "r")
        n_lines = len(file.read().splitlines())

        file.close()

        # Check that the file has the correct number of energies and zenith angles

        if n_lines == len(constants.ENERGIES) * len(angles):

            s_fluxes = np.reshape(
                np.loadtxt(file_name)[:, 2],
                (len(constants.ENERGIES), len(angles)),
            )

            if constants.get_verbose() > 1:

                print("Loaded surface fluxes.")

            return s_fluxes

        else:

            print(n_lines)
            print(len(constants.ENERGIES))
            print(len(angles))

            return no_file(force=force)

    else:

        return no_file(force=force)


# Read the energies and angles from a surface fluxes file


def print_s_fluxes_grids(file_name):

    """Return the surface energy grid and zenith angles in a surface fluxes file."""

    file_contents = np.loadtxt(
        os.path.join(constants.get_directory(), "surface_fluxes", file_name)
    )

    file_s_energies = np.unique(file_contents[:, 0])
    file_angles = np.unique(file_contents[:, 1])

    print("This file has " + str(len(file_s_energies)) + " surface energies:")
    print(file_s_energies)
    print("This file has " + str(len(file_angles)) + " zenith angles:")
    print(file_angles)
